-- Word Generator for Traveller
-- Jeff Zeitlin (jeff.zeitlin@execnet.com)

-- Generates words using the Traveller word generation rules.  The language
-- rules are declared in a file structured like a Windows INI file; that is,
-- divided into sections headed by a [HEADER] line, followed by a series of
-- key=value lines.  See WORDGEN.TXT for more information.

-- These include files are distributed with the Euphoria system
include wildcard.e
include file.e
include get.e
include sort.e
              
-- These include files (and others) are included with the Wordgen
-- distribution as "jz.zip"              
include jz\assoc.e
include jz\stddefs.e
include jz\cmdline.e

assoc_list tables
tables = {}

sequence table_name
table_name = ""

procedure ProcessLine(sequence line)
    sequence table_value
    sequence key
    sequence val
    integer split
    sequence valget
    
    if length(line) = 0 then    -- blank line, ignore
        return
    elsif line[1] = ';' then    -- comment, ignore
        return
    elsif line[1] = '[' then    -- section start
        table_name = upper(line[2..length(line)-1])
        table_value = alist_reference(tables, table_name)
        tables = alist_set(tables,table_name,table_value)
    else                        -- presumed key=value line
        split = find('=',line)
        key = line[1..split-1]
        val = line[split+1..length(line)]
        table_value = alist_reference(tables,table_name)
        if (compare(table_name,"SYLLABLES") = 0) or 
           (compare(table_name,"ASSIMILATE") = 0) then
            table_value = alist_set(table_value,key,val)
        else
            valget = value(val)
            if valget[1] != GET_SUCCESS then
                puts(STD_ERR,"Non-numeric frequency of occurrence.\n")
                abort(1)
            end if
            table_value = table_value & repeat(key,valget[2])
        end if
        tables = alist_set(tables,table_name,table_value)
    end if
end procedure

sequence lang

procedure LoadTables()
    integer infile
    object inline
    
    infile = open(lang,"r")
    if infile = -1 then
        puts(STD_ERR,"Can't open " & lang & " for reading.\n")
        abort(1)
    end if
    inline = gets(infile)
    while not atom(inline) do
        inline = trim(inline)
        ProcessLine(inline)
        inline = gets(infile)
    end while
    close(infile)
end procedure

function MakeWord()
    sequence word
    sequence syltype
    object syllables
    sequence usetable
    integer pos
    object key
    object val

    word = ""
    usetable = alist_reference(tables,"LENGTH")
    if length(usetable)=0 then
        syllables = rand(6)
    else
        syllables = value(usetable[rand(length(usetable))])
        if syllables[1] = 0 then
            syllables = syllables[2]
        else
            puts(STD_ERR,"Invalid value for expected numeric syl. count\n")
            abort(1)
        end if
    end if
    syltype = "FIRST"
    for i = 1 to syllables do
        usetable = alist_reference(tables,"SYLLABLES")
        usetable = alist_reference(usetable,syltype)
        usetable = alist_reference(tables,usetable)
        if (length(usetable) = 0) then
            return word
        end if
        syltype = usetable[rand(length(usetable))]
        if syltype[1] = 'C' then
            usetable = alist_reference(tables,"INITIAL")
            word = word & usetable[rand(length(usetable))]
        end if
        usetable = alist_reference(tables,"VOWEL")
        word = word & usetable[rand(length(usetable))]
        if syltype[length(syltype)] = 'C' then
            usetable = alist_reference(tables,"FINAL")
            word = word & usetable[rand(length(usetable))]
        end if
    end for
    usetable = alist_reference(tables,"ASSIMILATE")
    if (length(usetable) > 0) then
        usetable = alist_iterator(usetable)
        for i = 1 to length(usetable) do
            key = usetable[i][1]
            val = usetable[i][2]
            pos = match(key,word)
            if pos != 0 then
                word = word[1..pos-1] & val & word[pos+length(key)..length(word)]
            end if
        end for
    end if
    return word
end function

integer nwords
sequence destname

procedure GetParams()
    object val
    sequence nval
    
    lang = {}
    val = switchval("l",bTRUE)
    if atom(val) then
        val = switchval("L",bTRUE)
        if atom(val) then
            val = "wordgen"
        end if
    end if
    lang = val & ".lng"
    destname = val & ".txt"
    nwords = 0
    val = switchval("n",bTRUE)
    if atom(val) then
        val = switchval("N",bTRUE)
        if atom(val) then
            val = "100"
        end if
    end if
    nval=value(val)
    if nval[1]=GET_SUCCESS then
        nwords = nval[2]
    else
        puts(STD_ERR, "Expected -n parameter not numeric\n")
        abort(1)
    end if
    val = switchval("o",bTRUE)
    if atom(val) then
        val = switchval("O",bTRUE)
        if not atom(val) then
            destname = val
        else
            val = destname
        end if
    end if
    destname = val
end procedure

sequence wordlist

procedure CleanUp()
    integer dirty
    
    wordlist = sort(wordlist)
    dirty = 1
    while dirty do
        dirty = 0
        for i = 2 to length(wordlist) do
            if compare(wordlist[i],wordlist[i-1]) = 0 then
                wordlist = wordlist[1..i-1] & wordlist[i+1..length(wordlist)]
                wordlist = append(wordlist,MakeWord())
                dirty = 1
            end if
        end for
        wordlist = sort(wordlist)
    end while
end procedure

procedure WriteWords()
    integer outfile
    
    outfile = open(destname,"a")
    if outfile = -1 then
        puts(STD_ERR,"Can't open " & destname & " for writing.\n")
        abort(1)
    end if
    for i = 1 to length(wordlist) do
        puts(outfile,wordlist[i]&"\n")
    end for
    close(outfile)
end procedure

wordlist = {}

GetParams()
LoadTables()
for i = 1 to nwords do
    wordlist = append(wordlist, MakeWord())
end for
CleanUp()
WriteWords()

